/*
 * Copyright (C) 2014 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#ifndef __TRTL_USER_H__
#define  __TRTL_USER_H__
/** @file mock-turtle.h */

#ifndef __KERNEL__
#include <string.h>
#endif

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(_a) (sizeof(_a) / sizeof(_a[0]))
#endif


#define __TRTL_MSG_ID_MAX 128
#define __TRTL_MSG_ID_MAX_TRTL 32
#define __TRTL_MSG_ID_MAX_USER (__TRTL_MSG_ID_MAX - __TRTL_MSG_ID_MAX_TRTL)


/**
 * It lists all notification's code.
 */
enum trtl_cpu_notification {
	TRTL_CPU_NOTIFY_APPLICATION = 0,
	__TRTL_CPU_NOTIFY_MAX,
};


/**
 * It enumerates the Mock Turtle message ID.
 * These IDs starts after the user defined ones
 */
enum trtl_msg_id {
	TRTL_MSG_ID_UNDEFINED = __TRTL_MSG_ID_MAX_USER,
	TRTL_MSG_ID_ERROR,
	TRTL_MSG_ID_PING,
	TRTL_MSG_ID_VER,
	TRTL_MSG_ID_DBG,
	TRTL_MSG_ID_VAR_GET,
	TRTL_MSG_ID_BUF_GET,
	TRTL_MSG_ID_VAR_SET,
	TRTL_MSG_ID_BUF_SET,
};


/**
 * Synchronous. When set, the sync_id is valid and the receiver
 * must answer with a message with the same sync_id
 */
#define TRTL_HMQ_HEADER_FLAG_SYNC (1 << 0)

/**
 * Acknowledgment. When set, the sync_id is valid and it is the
 * same as the sync message previously sent
 */
#define TRTL_HMQ_HEADER_FLAG_ACK  (1 << 1)


/**
 * Remote Procedure Call. when set the message ID will be used
 * to execute a special function by the receiver
 */
#define TRTL_HMQ_HEADER_FLAG_RPC  (1 << 2)


/**
 * HMQ header descriptor
 */
struct trtl_hmq_header {
	/* word 0 */
	uint16_t rt_app_id; /**< Real-Time application unique identifier */
	uint8_t flags;
	uint8_t msg_id; /**< Message identifier */
	/* word 1 */
	uint16_t len; /**< message-length */
	uint16_t sync_id;/**< synchronous identifier */
	/* word 2 */
	uint32_t seq; /**< sequence number (automatically set by the library) */
};


#define RT_VERSION_MAJ(_v) ((_v >> 16) & 0xFFFF)
#define RT_VERSION_MIN(_v) (_v & 0xFFFF)
#define RT_VERSION(_a, _b) (((_a & 0xFFFF) << 16) | (_b & 0xFFFF))

/**
 * It describes the version running on the embedded CPU
 */
struct trtl_fw_version {
	uint32_t rt_id; /**< RT application identifier */
	uint32_t rt_version; /**< RT application version*/
	uint32_t git_version; /**< git commit SHA1 of the compilation time */
};


/**
 * Maximum number of CPU core in a bitstream
 */
#define TRTL_MAX_CPU 8


/**
 * Maximum number of (H|R)MQ slots in a bitstream
 */
#define TRTL_MAX_MQ_CHAN 8


/**
 * Maximum number of Mock Turtle components on a single computer
 */
#define TRTL_MAX_CARRIER 20


/**
 * TLV used to embed buffers within a message
 */
struct trtl_tlv {
	uint32_t type; /**< buffer type */
	size_t size; /**< buffer size in byte */
	void *buf; /**< pointer to the buffer */
};


/**
 * Maximum number of notifications ID in the driver buffer
 */
#define TRTL_CPU_NOTIFY_HISTORY_N 128

/**
 * This value represents the maximum payload size (in words) of a single
 * message. The true maximum size is set in the internal ROM, but to simplify
 * the code here we set this to the maximum addressable payload: 4096 bytes
 */
#define TRTL_MAX_PAYLOAD_SIZE (4096 / 4)


/**
 * @enum trtl_smem_modifier
 * Shared memory operation modifier. This is a list of operation modifier
 * that you can use to access the shared memory.
 */
enum trtl_smem_modifier {
	TRTL_SMEM_TYPE_BASE = 0, /**< no operation */
	TRTL_SMEM_TYPE_ADD, /**< atomic addition */
	TRTL_SMEM_TYPE_SUB, /**< atomic subtraction */
	TRTL_SMEM_TYPE_SET, /**< atomic bit set */
	TRTL_SMEM_TYPE_CLR, /**< atomic bit clear */
	TRTL_SMEM_TYPE_FLP, /**< atomic bit flip */
	TRTL_SMEM_TYPE_TST_SET, /**< atomic test and set */
};


/**
 * @enum trtl_msg_filter_operation_type
 * List of available filter's operations
 */
enum trtl_msg_filter_operation_type {
	TRTL_MSG_FILTER_AND,
	TRTL_MSG_FILTER_OR,
	TRTL_MSG_FILTER_EQ,
	TRTL_MSG_FILTER_NEQ,
	__TRTL_MSG_FILTER_MAX,
};

#define TRTL_MSG_FILTER_FLAG_PAYLOAD (1 << 0)

/**
 * It describe a filter to apply to messages
 */
struct trtl_msg_filter {
	unsigned long flags; /**< options for the filter */
	uint32_t operation; /**< kind of operation to perform */
	uint32_t word_offset; /**< offset of the word to check */
	uint32_t mask; /**< mask to apply before the operation */
	uint32_t value; /**< second operand of the operation */
};


/**
 * Descriptor of the IO operation on Shared Memory
 */
struct trtl_smem_io {
	uint32_t addr; /**< address to access */
	uint32_t value; /**< value to write. After ioctl it will be overwritte
			   with the new value in the shared memory */
	int is_input; /**< flag to determinte data direction */
	enum trtl_smem_modifier mod;  /**< the kind of operation to do */
};

/**
 * @enum trtl_ioctl_commands
 * Available ioctl() messages
 */
enum trtl_ioctl_commands {
	TRTL_SMEM_IO, /**< access to shared memory */
	TRTL_MSG_FILTER_ADD, /**< add a message filter */
	TRTL_MSG_FILTER_CLEAN, /**< remove all filters */
};


#define TRTL_IOCTL_MAGIC 's'
#define TRTL_IOCTL_SMEM_IO _IOWR(TRTL_IOCTL_MAGIC, TRTL_SMEM_IO, \
				    struct trtl_smem_io)

#define TRTL_IOCTL_MSG_FILTER_ADD _IOW(TRTL_IOCTL_MAGIC,	\
				       TRTL_MSG_FILTER_ADD,	\
				       struct trtl_msg_filter)
#define TRTL_IOCTL_MSG_FILTER_CLEAN _IOW(TRTL_IOCTL_MAGIC,		\
					 TRTL_MSG_FILTER_CLEAN,		\
					 struct trtl_msg_filter)


/**
 * Messages descriptor
 */
struct trtl_msg {
#ifdef __TRTL_FIRMWARE__
	/* Only firmware */
	struct trtl_hmq_header *header;
	void *payload;
#else
	/* library and kernel */
	struct trtl_hmq_header hdr;
	uint32_t data[TRTL_MAX_PAYLOAD_SIZE]; /**< payload very maximum size */
#endif
};


#define TRTL_CONFIG_ROM_MQ_ENTRIES_SHIFT 16
#define TRTL_CONFIG_ROM_MQ_ENTRIES_MASK 0x00FF0000
#define TRTL_CONFIG_ROM_MQ_PAYLOAD_SHIFT 8
#define TRTL_CONFIG_ROM_MQ_PAYLOAD_MASK 0x0000FF00
#define TRTL_CONFIG_ROM_MQ_HEADER_SHIFT 0
#define TRTL_CONFIG_ROM_MQ_HEADER_MASK 0x000000FF

#define TRTL_CONFIG_ROM_MQ_SIZE_ENTRIES(_size) (1 << ((_size & TRTL_CONFIG_ROM_MQ_ENTRIES_MASK) >> TRTL_CONFIG_ROM_MQ_ENTRIES_SHIFT))
#define TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD(_size) (1 << ((_size & TRTL_CONFIG_ROM_MQ_PAYLOAD_MASK) >> TRTL_CONFIG_ROM_MQ_PAYLOAD_SHIFT))
#define TRTL_CONFIG_ROM_MQ_SIZE_HEADER(_size) (1 << ((_size & TRTL_CONFIG_ROM_MQ_HEADER_MASK) >> TRTL_CONFIG_ROM_MQ_HEADER_SHIFT))

/**
 * The synthesis configuration for a single MQ.
 * Note that there is always an input and output channel for each declaration.
 */
struct trtl_config_rom_mq {
	uint32_t sizes;
	uint32_t endpoint_id;
};


/**
 * This signature must be present on all the configuration rom
 */
#define TRTL_CONFIG_ROM_SIGNATURE 0x5452544c


/**
 * The synthesis configuration ROM descriptor shows useful configuration
 * options during synthesis.
 */
struct trtl_config_rom {
	uint32_t signature; /**< we expect to see a known value */
	uint32_t version; /**< Mock Turtle version */
	uint32_t pad1[1];
	uint32_t clock_freq;
	uint32_t flags;
	uint32_t app_id; /**< Application ID */
	uint32_t n_cpu; /**< number of CPU */
	uint32_t smem_size; /**< shared memory size */
	uint32_t mem_size[TRTL_MAX_CPU]; /**< memory size for each CPU */
	uint32_t n_hmq[TRTL_MAX_CPU]; /**< number of HMQ for each CPU */
	uint32_t n_rmq[TRTL_MAX_CPU]; /**< number of RMQ for each CPU */
	uint32_t pad2[96];
	struct trtl_config_rom_mq hmq[TRTL_MAX_CPU][TRTL_MAX_MQ_CHAN];
	struct trtl_config_rom_mq rmq[TRTL_MAX_CPU][TRTL_MAX_MQ_CHAN];
};


#endif
