"""
@package docstring
@author: Federico Vaga <federico.vaga@cern.ch>
@copyright: Copyright (c) 2015 CERN
@license: LGPL 3 or later

SPDX-License-Identifier: LGPL-3.0-or-later
"""

from ctypes import *
import errno
import os

TRTL_CONFIG_ROM_MAX_CPU = 8
TRTL_CONFIG_ROM_MAX_HMQ = 8
TRTL_CONFIG_ROM_MAX_RMQ = 8


class TrtlConfigHmq(Structure):
    _fields_ = [
        ("sizes", c_uint32),
        ("endpoint_id", c_uint32),
    ]

    @property
    def entries(self):
        return 1 << ((self.sizes >> 16) & 0xFF)

    @property
    def payload_size(self):
        return 1 << ((self.sizes >> 8) & 0xFF)

    @property
    def header_size(self):
        return 1 << ((self.sizes >> 0) & 0xFF)



class TrtlConfig(Structure):
    TRTL_CONFIG_ROM_SIGNATURE = 0x5452544c
    _fields_ = [
        ("signature", c_uint32),
        ("version", c_uint32),
        ("pad", c_uint32),
        ("clock_freq", c_uint32),
        ("flags", c_uint32),
        ("app_id", c_uint32),
        ("n_cpu", c_uint32),
        ("smem_size", c_uint32),
        ("mem_size", c_uint32 * TRTL_CONFIG_ROM_MAX_CPU),
        ("n_hmq", c_uint32 * TRTL_CONFIG_ROM_MAX_CPU),
        ("n_rmq", c_uint32 * TRTL_CONFIG_ROM_MAX_CPU),
        ("pad2", c_uint32 * 96),
        ("hmq", TRTL_CONFIG_ROM_MAX_CPU * (TrtlConfigHmq * TRTL_CONFIG_ROM_MAX_HMQ)),
        ("rmq", TRTL_CONFIG_ROM_MAX_CPU * (TrtlConfigHmq * TRTL_CONFIG_ROM_MAX_HMQ)),
    ]

class TrtlMessageInternal(Structure):
    _fields_ = [
        ("datalen", c_uint32),
        ("data", c_uint32 * 128),
        ("max_size", c_uint32),
        ("offset", c_uint32),
        ("direction", c_int),
        ("error", c_int),
    ]

class TrtlHmqHeader(Structure):
    TRTL_HMQ_HEADER_FLAG_SYNC=(1 << 0)
    _fields_ = [
        ("rt_app_id", c_uint16),
        ("flags",  c_uint8),
        ("msg_id", c_uint8),
        ("len", c_uint16),
        ("sync_id", c_uint16),
        ("seq",  c_uint32),
    ]

    def __eq__(self, other):
        if self.rt_app_id != other.rt_app_id:
            return False
        if self.flags != other.flags:
            return False
        if self.msg_id != other.msg_id:
            return False
        if self.len != other.len:
            return False
        if self.sync_id != other.sync_id:
            return False
        if self.seq != other.seq:
            return False
        return True

class TrtlMessage(Structure):
    _fields_ = [
        ("header", TrtlHmqHeader),
        ("payload", (c_uint32 * (4096))),
    ]

    def __eq__(self, other):
        if self.header != other.header:
            return False
        cnt = 0
        for v1, v2 in zip(self.payload, other.payload):
            print("{:x} {:x}".format(v1, v2))
            if v1 != v2:
                return False
            cnt += 1
            if cnt >= self.header.len:
                break
        return True


class TrtlFirmwareVersion(Structure):
    _fields_ = [
        ("id", c_uint32),
        ("fw-version", c_uint32),
        ("git-version", c_uint32),
    ]


class TrtlDevice(object):
    """It is a Python class that represent a Mock Turtle Device
    """

    def __init__(self, devid):
        if devid is None:
            raise Exception("Invalid device ID")
        self.__init_library()

        self.device_id = devid
        self.libtrtl.trtl_init()
        set_errno(0)
        self.tkn = self.libtrtl.trtl_open_by_id(self.device_id)
        rom_ptr = self.libtrtl.trtl_config_get(self.tkn)
        self.rom = rom_ptr.contents

        self.cpu = []
        for i in range(self.rom.n_cpu):
            self.cpu.append(TrtlCpu(self, i))
        self.shm = TrtlSmem(self)

    def __del__(self):
        self.libtrtl.close(self.tkn)
        self.libtrtl.trtl_exit()

    def __str__(self):
        return "trtl-{:04x}".format(self.device_id)

    def __repr__(self):
        return "{:s}".format(str(self))

    def __init_library(self):
        self.libtrtl = CDLL("libmockturtle.so", use_errno=True)

        self.libtrtl.trtl_open_by_id.argtypes = [c_uint]
        self.libtrtl.trtl_open_by_id.restype = c_void_p
        self.libtrtl.trtl_open_by_id.errcheck = self.__errcheck_pointer
        # ERROR
        self.libtrtl.trtl_strerror.argtypes = [c_int]
        self.libtrtl.trtl_strerror.restype = c_char_p
        self.libtrtl.trtl_open_by_id.errcheck = self.__errcheck_pointer
        # SHM READ
        self.libtrtl.trtl_smem_read.argtypes = [c_void_p, c_uint,
                                                POINTER(c_int), c_size_t,
                                                c_uint]
        self.libtrtl.trtl_smem_read.restype = c_int
        self.libtrtl.trtl_smem_read.errcheck = self.errcheck
        # SHM WRITE
        self.libtrtl.trtl_smem_write.argtypes = [c_void_p, c_uint,
                                                 POINTER(c_int), c_size_t,
                                                 c_uint]
        self.libtrtl.trtl_smem_write.restype = c_int
        self.libtrtl.trtl_smem_write.errcheck = self.errcheck
        # CPU Enable
        self.libtrtl.trtl_cpu_enable.argtypes = [c_void_p, c_uint]
        self.libtrtl.trtl_cpu_enable.restype = c_int
        self.libtrtl.trtl_cpu_enable.errcheck = self.errcheck
        # CPU Disable
        self.libtrtl.trtl_cpu_disable.argtypes = [c_void_p, c_uint]
        self.libtrtl.trtl_cpu_disable.restype = c_int
        self.libtrtl.trtl_cpu_disable.errcheck = self.errcheck
        # CPU Is Enable
        self.libtrtl.trtl_cpu_is_enable.argstypes = [c_void_p,
                                                     c_uint,
                                                     POINTER(c_int)]
        self.libtrtl.trtl_cpu_is_enable.restype = c_int
        self.libtrtl.trtl_cpu_is_enable.errcheck = self.errcheck
        # CPU Load
        self.libtrtl.trtl_cpu_load_application_file.argtypes = [c_void_p,
                                                                c_uint,
                                                                c_char_p]
        self.libtrtl.trtl_cpu_load_application_file.restype = c_int
        self.libtrtl.trtl_cpu_load_application_file.errcheck = self.errcheck
        # CPU Dump
        self.libtrtl.trtl_cpu_dump_application_file.argtypes = [c_void_p,
                                                                c_uint,
                                                                c_char_p]
        self.libtrtl.trtl_cpu_dump_application_file.restype = c_int
        self.libtrtl.trtl_cpu_dump_application_file.errcheck = self.errcheck
        # GET CONFIG
        self.libtrtl.trtl_config_get.argtypes = [c_void_p, ]
        self.libtrtl.trtl_config_get.restype = POINTER(TrtlConfig)
        self.libtrtl.trtl_config_get.errcheck = self.__errcheck_pointer
        # ASYNC RECEIVE
        self.libtrtl.trtl_msg_async_recv.argtypes = [c_void_p,
                                                     c_uint,
                                                     c_uint,
                                                     POINTER(TrtlMessage),
                                                     c_uint,
                                                    ]
        self.libtrtl.trtl_msg_async_recv.restype = c_int
        self.libtrtl.trtl_msg_async_recv.errcheck = self.errcheck
        # ASYNC SEND
        self.libtrtl.trtl_msg_async_send.argtypes = [c_void_p,
                                                     c_uint,
                                                     c_uint,
                                                     POINTER(TrtlMessage),
                                                     c_uint,
                                                    ]
        self.libtrtl.trtl_msg_async_send.restype = c_int
        self.libtrtl.trtl_msg_async_send.errcheck = self.errcheck
        # SYNC SEND and RECEIVE
        self.libtrtl.trtl_msg_sync.argtypes = [c_void_p,
                                               c_uint,
                                               c_uint,
                                               POINTER(TrtlMessage),
                                               POINTER(TrtlMessage),
                                               c_uint,
                                              ]
        self.libtrtl.trtl_msg_sync.restype = c_int
        self.libtrtl.trtl_msg_sync.errcheck = self.errcheck
        # FLUSH
        self.libtrtl.trtl_hmq_flush.argtypes = [c_void_p,
                                                c_uint,
                                                c_uint,
                                               ]
        self.libtrtl.trtl_hmq_flush.restype = c_int
        self.libtrtl.trtl_hmq_flush.errcheck = self.errcheck
        # FW PING
        self.libtrtl.trtl_fw_ping.argtypes = [c_void_p, c_uint, c_uint]
        self.libtrtl.trtl_fw_ping.restype = c_int
        self.libtrtl.trtl_fw_ping.errcheck = self.__errcheck_int
        # FW VERSION
        self.libtrtl.trtl_fw_version.argtypes = [c_void_p, c_uint, c_uint,
                                                 POINTER(TrtlFirmwareVersion),
                                                ]
        self.libtrtl.trtl_fw_version.restype = c_int
        self.libtrtl.trtl_fw_version.errcheck = self.__errcheck_int
        # self.libtrtl.trtl_hmq_filter_add.argtypes = [c_void_p]
        # self.libtrtl.trtl_hmq_filter_clean.argtypes = [c_void_p]
        # # Return
        # self.libtrtl.trtl_hmq_filter_add.restype = c_int
        # self.libtrtl.trtl_hmq_filter_clean.restype = c_int
        # # Error
        # self.libtrtl.trtl_hmq_filter_add.errcheck = self.errcheck
        # self.libtrtl.trtl_hmq_filter_clean.errcheck = self.errcheck

    def __errcheck_pointer(self, ret, func, args):
        """Generic error handler for functions returning pointers"""
        if ret is None:
             raise OSError(get_errno(),
                           self.libtrtl.trtl_strerror(get_errno()), "")
        else:
             return ret

    def __errcheck_int(self, ret, func, args):
        """Generic error checker for functions returning 0 as success
        and -1 as error"""
        if ret < 0:
            raise OSError(get_errno(),
                          self.libtrtl.trtl_strerror(get_errno()), "")
        else:
            return ret

    def errcheck(self, ret, func, args):
        self.__errcheck_int


class TrtlCpu(object):
    """
    Python wrapper for CPU management
    """

    def __init__(self, trtl_dev, idx_cpu):
        self.trtl_dev = trtl_dev
        self.idx_cpu = idx_cpu

        self.hmq = []
        for i in range(self.trtl_dev.rom.n_hmq[self.idx_cpu]):
            self.hmq.append(TrtlHmq(self, i))

        self.libtrtl = self.trtl_dev.libtrtl

    def __str__(self):
        return "{:s}-{:02d}".format(str(self.trtl_dev), self.idx_cpu)

    def __repr__(self):
        return "{:s}".format(str(self))

    def enable(self):
        """
        It enables a CPU; in other words, it clear the reset line of a CPU.
        @exception OSError from C library errors
        """
        self.libtrtl.trtl_cpu_enable(self.trtl_dev.tkn, self.idx_cpu)

    def disable(self):
        """
        It disables a CPU; in other words, it sets the reset line of a CPU.
        @exception OSError from C library errors
        """
        self.libtrtl.trtl_cpu_disable(self.trtl_dev.tkn, self.idx_cpu)

    def is_enable(self):
        """
        It checks if the CPU is enabled (or not)
        @return True when the CPU is enable, False otherwise
        @exception OSError from C library errors
        """
        enable = c_int(0)
        self.libtrtl.trtl_cpu_is_enable(self.trtl_dev.tkn, self.idx_cpu,
                                        byref(enable))
        return True if enable.value else False

    def load_application_file(self, file_path):
        """
        It loads a firmware from the given file
        @param[in] file_path path to the firmware file
        @exception OSError from C library errors
        """
        self.libtrtl.trtl_cpu_load_application_file(self.trtl_dev.tkn,
                                                    self.idx_cpu,
                                                    file_path.encode())

    def dump_application_file(self, file_path):
        """
        It dumps the running firmware to the given file
        @param[in] file_path path to the firmware file
        @exception OSError from C library errors
        """
        self.libtrtl.trtl_cpu_dump_application_file(self.trtl_dev.tkn,
                                                    self.idx_cpu,
                                                    file_path.encode())

    def ping(self, idx_hmq):
        """
        It pings the firmware running on the CPU
        @return True if the firmware is alive, False otherwise
        """
        val = self.libtrtl.trtl_fw_ping(self.trtl_dev.tkn,
                                        self.idx_cpu, idx_hmq)
        return True if val == 0 else False

    def version(self):
        version = TrtlFirmwareVersion()
        self.libtrtl.trtl_fw_version(self.trtl_dev.tkn,
                                     self.idx_cpu, idx_hmq,
                                     pointer(version))
        return version


class TrtlHmq(object):
    """
    Python wrapper for HMQ management
    """
    def __init__(self, trtl_cpu, idx_hmq):
        self.trtl_cpu = trtl_cpu
        self.trtl_dev = self.trtl_cpu.trtl_dev
        self.idx_cpu = self.trtl_cpu.idx_cpu
        self.idx_hmq = idx_hmq
        self.libtrtl = self.trtl_dev.libtrtl

    def __str__(self):
        return "{:s}-{:02d}".format(str(self.trtl_cpu), self.idx_hmq)

    def __repr__(self):
        return "{:s}".format(str(self))

    def flush(self):
        """It flushes the content of the HMQ for both input and output
        """
        set_errno(0)
        self.libtrtl.trtl_hmq_flush(self.trtl_dev.tkn,
                                    self.idx_cpu, self.idx_hmq)

    def send_msg(self, msg, timeout=-1):
        """
        It sends an asynchronous message
        @param[in] msg message to send
        @param[in] timeout time to wait before returning
        @exception OSError from C library errors
        """
        set_errno(0)
        self.libtrtl.trtl_msg_async_send(self.trtl_dev.tkn,
                                         self.idx_cpu, self.idx_hmq,
                                         pointer(msg), 1)

    def recv_msg(self, timeout=-1):
        """
        It receives an asynchronous message
        @param[in] timeout time to wait before returning
        @return the synchronous answer as list of 32bit words
        @exception OSError from C library errors
        """
        # TODO test me when you can
        set_errno(0)
        msg = TrtlMessage()
        self.libtrtl.trtl_msg_async_recv(self.trtl_dev.tkn,
                                         self.idx_cpu, self.idx_hmq,
                                         pointer(msg), 1)
        return msg

    def sync_msg(self, msg_s, timeout=1000):
        """
        It sends a synchronous message
        @param[in] hmq_out index of the HMQ output slot
        @param[in] msg_list list of 32bit words to send
        @param[in] timeout time to wait before returning
        @return the synchronous answer as a set of header and payload
        @exception OSError from C library errors
        """
        msg_r = TrtlMessage()
        self.libtrtl.trtl_msg_sync(self.trtl_dev.tkn,
                                   self.idx_cpu, self.idx_hmq,
                                   byref(msg_s), pointer(msg_r),
                                   timeout)
        return msg_r

class TrtlSmem(object):
    """
    Python wrapper for Shared Memory management
    """
    MOD_DIRECT = 0
    MOD_ADD = 1
    MOD_SUB = 2
    MOD_OR = 3
    MOD_CLR_AND = 4
    MOD_XOR = 5

    def __init__(self, trtl_dev):
        self.trtl_dev = trtl_dev

    def read(self, address, count):
        """
        It reads from the shared memory 'count' 32bit words starting
        from 'address'
        @param[in] address memory address where start writing
        @param[in] count number of 32bit words to read
        @return a list of integer
        @exception OSError from C library errors
        """
        dataArr = (c_int * count)()
        dataCas = cast(dataArr, POINTER(c_int))
        self.libtrtl.trtl_smem_read(self.trtl_dev.tkn, address, dataCas, count, 0)
        return list(dataArr)

    def write(self, address, values, modifer):
        """
        It writes 'values' to the shared memory starting from 'address'.
        @param[in] address memory address where start writing
        @param[in] values integer list to write
        @param[in] modifier write operation modifier. It changes the write
                   behaviour
        @exception OSError from C library errors
        """
        dataArr = (c_int * len(values))(*values)
        dataCas = cast(dataArr, POINTER(c_int))
        self.libtrtl.trtl_smem_write(self.trtl_dev.tkn, address, dataCas,
                                     len(values), modifer)

    def write_direct(self, address, values):
        self.write(address, values, self.MOD_DIRECT)

    def write_add(self, address, values):
        self.write(address, values, self.MOD_ADD)

    def write_sub(self, address, values):
        self.write(address, values, self.MOD_SUB)

    def write_or(self, address, values):
        self.write(address, values, self.MOD_OR)

    def write_clr_and(self, address, values):
        self.write(address, values, self.MOD_CLR_AND)

    def write_xor(self, address, values):
        self.write(address, values, self.MOD_XOR)
