/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *         inspired by a draft of Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#ifndef __LIB_TRTL_H__
#define __LIB_TRTL_H__
/** @file libmockturtle.h */

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h>
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <poll.h>
#include "mockturtle.h"

extern const unsigned int trtl_default_timeout_ms;

struct trtl_dev;

#define TRTL_SYSFS_PATH_LEN 128


/**
 * This structure mimic the pollfd structure for the Mock Turtle needs
 */
struct polltrtl {
	struct trtl_dev *trtl;
	unsigned int idx_cpu;
	unsigned int idx_hmq;
	short events; /**< like in pollfd poll(2) */
	short revents; /**< like in pollfd poll(2) */
};

/**
 * Debug descriptor. It is not obfuscated because it is meant for debugging
 * purpose. This way, it leaves the user all the freedom to read/poll the
 * debug channel as (s)he wants.
 */
struct trtl_dbg {
	struct trtl_dev *trtl; /**< token of the device */
	unsigned int cpu_index; /**< CPU where read debug messages */
	int fd; /**< file descriptor of the debug interface */
};


/**
 * HMQ slot descriptor
 * @deprecated it is not necessary anymore
 */
struct trtl_hmq {
	struct trtl_dev *trtl; /**< where this slot belong to */
	char syspath[TRTL_SYSFS_PATH_LEN];
	unsigned int index; /* index of the slot. Note that we have
			       different kind of slot and each kind start
			       counting from 0*/
	unsigned long flags; /**< flags associated to the slot */
	int fd; /**< file descriptor */
};

#define TRTL_NAME_LEN 16
#define TRTL_PATH_LEN 32

#define TRTL_SYSFS_READ_LEN 32
#define TRTL_DEVICE_PATH_LEN 64


#define TRTL_HMQ_INCOMING	(1 << 0)
#define TRTL_HMQ_OUTCOMING	0x0


/**
 * Error codes for Mock-Turtle applications
 */
enum trtl_error_number {
	ETRTL_INVAL_PARSE = 83630, /**< cannot parse data from sysfs */
	ETRTL_INVAL_SLOT, /**< invalid slot */
	ETRTL_NO_IMPLEMENTATION, /**< a prototype is not implemented */
	ETRTL_HMQ_CLOSE, /**< The HMQ is closed */
	ETRTL_INVALID_MESSAGE, /**< Invalid message */
	ETRTL_HMQ_READ, /**< Error while reading messages */
	ETRTL_MSG_SYNC_FAILED_SEND, /**< Send sync message failure */
	ETRTL_MSG_SYNC_FAILED_RECV, /**< Receive sync message failure */
	ETRTL_MSG_SYNC_FAILED_RECV_TIMEOUT, /**< Receive sync message failure:
					       timeout */
	ETRTL_MSG_SYNC_FAILED_INVAL, /**< Receive sync message failure:
					  invalid */
	__ETRTL_MAX,
};


/**
 * @file libmockturtle.c
 */
/**
 * @defgroup dev Device
 * Set of functions to perform basic openartion on the device
 * @{
 */
extern int trtl_init();
extern void trtl_exit();
extern uint32_t trtl_count();
extern char **trtl_list();
extern void trtl_list_free(char **list);

extern struct trtl_dev *trtl_open(const char *device);
extern struct trtl_dev *trtl_open_by_id(uint32_t device_id);
extern struct trtl_dev *trtl_open_by_lun(unsigned int lun) __attribute__((deprecated));
extern void trtl_close(struct trtl_dev *trtl);
extern char *trtl_name_get(struct trtl_dev *trtl);
const struct trtl_config_rom *trtl_config_get(struct trtl_dev *trtl);
/**@}*/

/**
 * @defgroup util Utilities
 * Utilities collection
 * @{
 */
extern void trtl_print_header(struct trtl_msg *msg);
extern void trtl_print_payload(struct trtl_msg *msg);
extern void trtl_print_message(struct trtl_msg *msg);
extern char *trtl_strerror(int err);
/**@}*/

/**
 * @defgroup cpu CPU
 * Set of function that allow you to manage the FPGA cores
 * @{
 */
extern int trtl_cpu_load_application_raw(struct trtl_dev *trtl,
					 unsigned int index,
					 void *code, size_t length,
					 unsigned int offset);
extern int trtl_cpu_load_application_file(struct trtl_dev *trtl,
					  unsigned int index,
					  char *path);
extern int trtl_cpu_dump_application_raw(struct trtl_dev *trtl,
					 unsigned int index,
					 void *code, size_t length,
					 unsigned int offset);
extern int trtl_cpu_dump_application_file(struct trtl_dev *trtl,
					  unsigned int index,
					  char *path);

extern int trtl_cpu_reset_set(struct trtl_dev *trtl, uint32_t mask);
extern int trtl_cpu_reset_get(struct trtl_dev *trtl, uint32_t *mask);
extern int trtl_cpu_enable(struct trtl_dev *trtl, unsigned int index);
extern int trtl_cpu_disable(struct trtl_dev *trtl, unsigned int index);
extern int trtl_cpu_is_enable(struct trtl_dev *trtl, unsigned int index,
			      unsigned int *enable);
/**@}*/

/**
 * @defgroup hmq Host Message Queue
 * Functions to manage the HMQ slots: configuration and transmission
 * @{
 */

/* FIXME to be tested */
extern int trtl_hmq_filter_add(struct trtl_dev *trtl,
			       unsigned int idx_cpu,
			       unsigned int idx_hmq,
			       struct trtl_msg_filter *filter);
/* FIXME to be tested */
extern int trtl_hmq_filter_clean(struct trtl_dev *trtl,
				 unsigned int idx_cpu,
				 unsigned int idx_hmq);
extern int trtl_bind(struct trtl_dev *trtl, struct trtl_msg_filter *flt,
		     unsigned int length);

extern int trtl_hmq_fd(struct trtl_dev *trtl,
		       unsigned int idx_cpu,
		       unsigned int idx_hmq);
extern int trtl_msg_async_send(struct trtl_dev *trtl,
			       unsigned int idx_cpu,
			       unsigned int idx_hmq,
			       struct trtl_msg *msg,
			       unsigned int n);
extern int trtl_msg_async_recv(struct trtl_dev *trtl,
			       unsigned int idx_cpu,
			       unsigned int idx_hmq,
			       struct trtl_msg *msg,
			       unsigned int n);
extern int trtl_msg_poll(struct polltrtl *trtlp,
			 unsigned int npolls, int timeout);
extern int trtl_msg_sync(struct trtl_dev *trtl,
			 unsigned int idx_cpu,
			 unsigned int idx_hmq,
			 struct trtl_msg *msg_s,
			 struct trtl_msg *msg_r,
			 int timeout);
extern int trtl_hmq_flush(struct trtl_dev *trtl,
			  unsigned int idx_cpu,
			  unsigned int idx_hmq);
/**@}*/

/**
 * @defgroup smem Shared Memory
 * Functions to access the shared memory from the host
 * @{
 */
extern int trtl_smem_read(struct trtl_dev *trtl, uint32_t addr, uint32_t *data,
			  size_t count, enum trtl_smem_modifier mod);
extern int trtl_smem_write(struct trtl_dev *trtl, uint32_t addr, uint32_t *data,
			   size_t count, enum trtl_smem_modifier mod);
/**@}*/


/**
 * @defgroup rtmsg Real Time service messages
 * Message builders for RT service messages
 * @{
 */
extern int trtl_fw_version(struct trtl_dev *trtl,
			   unsigned int idx_cpu,
			   unsigned int idx_hmq,
			   struct trtl_fw_version *version);
extern int trtl_fw_ping(struct trtl_dev *trtl,
			unsigned int idx_cpu,
			unsigned int idx_hmq);
extern int trtl_fw_variable_set(struct trtl_dev *trtl,
				unsigned int idx_cpu,
				unsigned int idx_hmq,
				uint32_t *var, unsigned int n_var);
extern int trtl_fw_variable_get(struct trtl_dev *trtl,
				unsigned int idx_cpu,
				unsigned int idx_hmq,
				uint32_t *var, unsigned int n_var);
extern int trtl_fw_buffer_set(struct trtl_dev *trtl,
			      unsigned int idx_cpu,
			      unsigned int idx_hmq,
			      struct trtl_tlv *tlv,
			      unsigned int n_tlv);
extern int trtl_fw_buffer_get(struct trtl_dev *trtl,
			      unsigned int idx_cpu,
			      unsigned int idx_hmq,
			      struct trtl_tlv *tlv,
			      unsigned int n_tlv);
/**@}*/
#ifdef __cplusplus
};
#endif

#endif
