/**
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <errno.h>
#include <inttypes.h>
#include <stdint.h>
#include <mockturtle-framework.h>

uint32_t msg_seq = 0;
struct trtl_fw_application *_app;

/**
 * It makes the given message an error message
 */
void trtl_fw_message_error(struct trtl_msg *msg, int err)
{
	msg->header->msg_id = TRTL_MSG_ID_ERROR;
	msg->header->len = 1;
	((uint32_t *)msg->payload)[0] = err;
}


/**
 * This is an @ref trtl_fw_action_t function type. It fills the payload
 * witht an answer for the ping message.
 */
static int rt_recv_ping(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	msg_o->header->msg_id = TRTL_MSG_ID_PING;
	return 0;
}

/**
 * This is an @ref trtl_fw_action_t function type. It fills the payload with
 * version information.
 */
static int rt_version_getter(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	msg_o->header->msg_id = TRTL_MSG_ID_VER;
	msg_o->header->len = sizeof(struct trtl_fw_version) / 4;
	memcpy(msg_o->payload, (void *)&_app->version,
	       sizeof(struct trtl_fw_version));
	return 0;
}


/**
 * This is an @ref trtl_fw_action_t function type. Accorind the message request,
 * it copies one of the declared data structure to the output payload.
 */
static int trtl_fw_buffer_getter(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	unsigned int offset = 0, index, size;
	uint32_t *din;
	uint32_t *dout;

	if (!HAS_MOCKTURTLE_FRAMEWORK_BUFFER_ENABLE) {
		return -EPERM;
	}

	/* Only synchronous */
	if (!(msg_i->header->flags & TRTL_HMQ_HEADER_FLAG_SYNC))
		return -EINVAL;

	msg_o->header->msg_id = TRTL_MSG_ID_BUF_GET;
	din = msg_i->payload;
	dout = msg_o->payload;

	while (offset < msg_i->header->len) {
		pr_debug("%s: offset %d/%d\n\r", __func__,
			 offset, msg_i->header->len);

		index = din[offset];
		dout[offset++] = index;
		size = din[offset];
		dout[offset++] = size;
		pr_debug("%s Type %d Len %d Addr 0x%p\n\r", __func__,
			  index, size, _app->buffers[index].buf);

		if (_app->buffers[index].len == size) {
			memcpy(&dout[offset],
			       (uint32_t *)_app->buffers[index].buf,
			       size);
		} else {
			pr_error("%s: structure %d len not correct %"PRId32" != %d\n\r",
				  __func__, index, _app->buffers[index].len, size);
		}

		offset += (size / 4); /* Next TLV record */
	}

	return 0;
}


/**
 * This is an @ref trtl_fw_action_t function type. Accorind the message request,
 * it overwrite a data structure with the one contained in the input
 * payload. If the message is syncrnous it will copy back the data structure.
 */
static int trtl_fw_buffer_setter(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	unsigned int offset = 0, index, size;
	uint32_t *din;

	if (!HAS_MOCKTURTLE_FRAMEWORK_BUFFER_ENABLE) {
		return -EPERM;
	}

	din = msg_i->payload;

	while (offset < msg_i->header->len) {
		pr_debug("%s: offset %d/%d\n\r", __func__,
			 offset, msg_i->header->len);

		index = din[offset++];
		size = din[offset++];

		pr_debug("%s Type %d Len %d Addr 0x%p\n\r", __func__,
			  index, size, _app->buffers[index].buf);

		if (_app->buffers[index].len == size) {
			memcpy((uint32_t *)_app->buffers[index].buf,
			       &din[offset], size);
		} else {
			pr_error("%s:%d structure %d len not correct %"PRId32" != %d\n\r",
				  __func__, __LINE__, index,
				  _app->buffers[index].len, size);
		}

		offset += (size / 4); /* Next TLV record */
	}

	/* Return back new values. Host can compare with what it sent
	   to spot errors */
	if (msg_i->header->flags & TRTL_HMQ_HEADER_FLAG_SYNC)
		return trtl_fw_buffer_getter(msg_i, msg_o);

	return 0;
}


/**
 * This is an @ref trtl_fw_action_t function type. Accorind the message request,
 * it copies a number of declared variables.
 */
static int trtl_fw_variable_getter(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	struct trtl_fw_variable *var;
	uint32_t *dout, *din , *mem, val;
	int i;

	if (!HAS_MOCKTURTLE_FRAMEWORK_VARIABLE_ENABLE) {
		return -EPERM;
	}

	/* Only synchronous */
	if (!(msg_i->header->flags & TRTL_HMQ_HEADER_FLAG_SYNC))
		return -EINVAL;

	/* we always have a pair of values  */
	if (msg_i->header->len & 0x1)
		return -EINVAL;

	din = msg_i->payload;
	dout = msg_o->payload;
	msg_o->header->msg_id = TRTL_MSG_ID_VAR_GET;
	msg_o->header->len = msg_i->header->len;

	/* Write all values in the proper place */
	for (i = 0; i < msg_o->header->len; i += 2) {
		if (din[i] >= _app->n_variables) {
			dout[i] = ~0; /* Report invalid index */
			continue;
		}
		dout[i] = din[i];
		var = &_app->variables[dout[i]];
		mem = (uint32_t *) var->addr;
		val = (*mem >> var->offset) & var->mask;
		dout[i + 1] = val;

		pr_debug("%s index %"PRIu32"d/%d | [0x%p] = 0x%08"PRIx32" -> 0x%08"PRIx32" | index in msg (%d/%d)\n\r",
			  __func__,
			  dout[i], _app->n_variables - 1,
			  mem, *mem,  dout[i + 1],
			  i + 1, msg_i->header->len - 1);
	}

	return 0;
}


/**
 * This is an @ref trtl_fw_action_t function type. Accorind the message request,
 * it writes a number of declared variables. If the message is synchronous
 * it copies back the values in the output payload.
 */
static int trtl_fw_variable_setter(struct trtl_msg *msg_i, struct trtl_msg *msg_o)
{
	struct trtl_fw_variable *var;
	uint32_t *din, *mem, val;
	int i;

	if (!HAS_MOCKTURTLE_FRAMEWORK_VARIABLE_ENABLE) {
		return -EPERM;
	}

	/* we always have a pair of values  */
	if (msg_i->header->len & 0x1)
		return -EINVAL;

	din = msg_i->payload;

	/* Write all values in the proper place */
	for (i = 0; i < msg_i->header->len; i += 2) {
		if (din[i] >= _app->n_variables)
			continue;
		var = &_app->variables[din[i]];
		mem = (uint32_t *) var->addr;
		val = ((din[i + 1] & var->mask) << var->offset);
		if (var->flags & TRTL_FW_VARIABLE_FLAG_FLD)
			*mem = (*mem & ~var->mask) | val;
		else
			*mem = val;

		pr_debug("%s index %"PRIu32"/%d | [0x%p] = 0x%08"PRIx32" <- 0x%08"PRIx32" (0x%08"PRIx32") | index in msg (%d/%d)\n\r",
			  __func__,
			  din[i], _app->n_variables - 1,
			  mem, *mem, val, din[i + 1],
			  i + 1, msg_i->header->len - 1);
	}

	/* Return back new values. Host can compare with what it sent
	   to spot errors */
	if (msg_i->header->flags & TRTL_HMQ_HEADER_FLAG_SYNC)
		return trtl_fw_variable_getter(msg_i, msg_o);

	return 0;
}


/**
 * List of standard actions
 */
static trtl_fw_action_t *trtl_actions_in[] = {
	[TRTL_MSG_ID_PING - __TRTL_MSG_ID_MAX_USER] = rt_recv_ping,
	[TRTL_MSG_ID_VER - __TRTL_MSG_ID_MAX_USER] = rt_version_getter,
	[TRTL_MSG_ID_VAR_SET - __TRTL_MSG_ID_MAX_USER] = trtl_fw_variable_setter,
	[TRTL_MSG_ID_VAR_GET - __TRTL_MSG_ID_MAX_USER] = trtl_fw_variable_getter,
	[TRTL_MSG_ID_BUF_SET - __TRTL_MSG_ID_MAX_USER] = trtl_fw_buffer_setter,
	[TRTL_MSG_ID_BUF_GET - __TRTL_MSG_ID_MAX_USER] = trtl_fw_buffer_getter,
};


/**
 * According to the given message ID it returns the correspondent action
 * @param[in] msg_id message id
 * @return pointer to action function
 */
static inline trtl_fw_action_t *rt_action_get(unsigned int msg_id)
{
	unsigned int idx_act = msg_id;

	if (idx_act < __TRTL_MSG_ID_MAX_USER) {
		/* User Actions */
		if (idx_act >= _app->n_actions || !_app->actions[idx_act]) {
			pr_error("Unknown action (usr) msg_id 0x%x\n\r",
				 msg_id);
			return NULL;
		}
		return _app->actions[idx_act];
	} else {
		/* Mock Turtle Standard Actions */
		idx_act -= __TRTL_MSG_ID_MAX_USER;
		if (ARRAY_SIZE(trtl_actions_in) < idx_act) {
			pr_error("Unknown action msg_id 0x%x\n\r",
				 msg_id);
			return NULL;
		}
		return trtl_actions_in[idx_act];
	}
}

/**
 * It runs the action associated with the given identifier
 * @param[in] id action identifier
 * @param[in] msg_i input message for the action
 * @param[out] msg_o output message from the action
 * @return 0 on success. Otherwise an error code
 *
 * If the action generates an error, the framework will send
 * an error message with the error code.
 */
static inline int rt_action_run(enum trtl_mq_type type,
				unsigned int idx_mq,
				struct trtl_msg *msg)
{
	struct trtl_msg msg_out;
	trtl_fw_action_t *action = rt_action_get(msg->header->msg_id);
	int err = 0;

	if (!action)
		return EINVAL;

	if (!(msg->header->flags & TRTL_HMQ_HEADER_FLAG_SYNC)) {
		/* Asynchronous message, then no output */
		return action(msg, NULL);
	}

	/* Synchronous message */
	mq_claim(type, idx_mq);
	mq_map_out_message(type, idx_mq, &msg_out);

	/* prepare the header */
	msg_out.header->rt_app_id = _app->version.rt_id;
	msg_out.header->flags = TRTL_HMQ_HEADER_FLAG_ACK;
	msg_out.header->msg_id = 0;
	msg_out.header->len = 0;
	msg_out.header->sync_id = msg->header->sync_id;
	msg_out.header->seq = _app->seq++;

	err = action(msg, &msg_out);
	if (err)
		trtl_fw_message_error(&msg_out, err);

	trtl_fw_print_message(&msg_out);

	mq_send(type, idx_mq);

	return err;
}


/**
 * It dispatch messages coming from a given message queue.
 * @param[in] mq_in MQ index within mq declaration in rt_application
 * @todo provide support for remote queue
 * @return 0 on success. -1 on error
 */
int trtl_fw_mq_action_dispatch(enum trtl_mq_type type, unsigned int idx_mq)
{
	struct trtl_msg msg;
	int err = 0;

	if (type == TRTL_RMQ)
		return -EPERM; /* not supported yet */

	/* Check if we have a pending message */
	if (!(mq_poll_in(type) & (1 << idx_mq)))
		return -EAGAIN;

	/* Map the message  */
	mq_map_in_message(type, idx_mq, &msg);

	trtl_fw_print_message(&msg);

	if (!(msg.header->flags & TRTL_HMQ_HEADER_FLAG_RPC)) {
		return -EINVAL;
	}

	if (msg.header->rt_app_id &&
	    msg.header->rt_app_id != _app->version.rt_id) {
		pr_error("Invalid app id 0x%x\n\r",
			 msg.header->rt_app_id);
		err = -EINVAL;
		goto out;
	}

	/* Run the correspondent action */
	err = rt_action_run(type, idx_mq, &msg);
	if (err)
		pr_error("Action failure err: %d\n\r", err);
out:
	mq_discard(type, idx_mq);
	return err;
}


/**
 * It builds and it sends a message over MQ.
 * The vargs will be copied into the payload message.
 * @param[in] type MQ type
 * @param[in] idx_mq MQ index within the declaration
 * @param[in] msg_id message identifier
 * @param[in] n number of vargs
 */
int trtl_fw_mq_send_uint32(enum trtl_mq_type type,
			   unsigned int idx_mq,
			   uint8_t msg_id,
			   unsigned int n, ...)
{
	struct trtl_msg msg;
	uint32_t sizes;
	va_list ap;
	int i;

	sizes = _app->cfgrom->hmq[trtl_get_core_id()][idx_mq].sizes;
	if (n > TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD(sizes))
		return -EINVAL;

	mq_claim(type, idx_mq);

	mq_map_out_message(type, idx_mq, &msg);
	msg.header->rt_app_id = _app->version.rt_id;
	msg.header->flags = 0;
	msg.header->msg_id = msg_id;
	msg.header->len = n;
	msg.header->seq = _app->seq++;

	va_start(ap, n);
	for (i = 0; i < msg.header->len;++i)
		((uint32_t *)msg.payload)[i] = va_arg(ap, uint32_t);
	va_end(ap);

	mq_send(type, idx_mq);

	return 0;
}


/**
 * It builds and it sends a message over MQ.
 * The buffer will be copied into the payload message.
 * Beware that, internally, it uses trtl_fw_mq_send().
 * @param[in] type MQ type
 * @param[in] idx_mq MQ index within the declaration
 * @param[in] msg_id message identifier
 * @param[in] n buffer size in bytes
 * @param[in] data buffer to send
 */
int trtl_fw_mq_send_buf(enum trtl_mq_type type,
			unsigned int idx_mq,
			uint8_t msg_id,
			unsigned int n,
			void *data)
{
	struct trtl_msg msg;
	uint32_t sizes;

	sizes = _app->cfgrom->hmq[trtl_get_core_id()][idx_mq].sizes;
	if ((n / 4) > TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD(sizes))
		return -EINVAL;

	mq_claim(type, idx_mq);

	mq_map_out_message(type, idx_mq, &msg);
	msg.header->rt_app_id = _app->version.rt_id;
	msg.header->flags = 0;
	msg.header->msg_id = msg_id;
	msg.header->len = n / 4;
	msg.header->seq = _app->seq++;
	memcpy(msg.payload, data, n);

	mq_send(type, idx_mq);

	return 0;
}


/**
 * It prints on the serial interface the given message header
 * @param[in] h message header to print
 */
void trtl_fw_print_header(struct trtl_hmq_header *h)
{
	if (!HAS_MOCKTURTLE_FRAMEWORK_DEBUG_ENABLE) {
		return;
	}

	pr_debug("\tapp_id 0x%04"PRIx16" | flags   0x%02"PRIx8" | msg_id 0x%02"PRIx8"\r\n",
		 h->rt_app_id, h->flags, h->msg_id);
	delay(1000);
	pr_debug("\tlen 0x%04"PRIx16" | sync_id 0x%04"PRIx16"\r\n",
		 h->len, h->sync_id);
	delay(1000);
	pr_debug("\tseq 0x%08"PRIx32"\r\n",
		 h->seq);
	delay(1000);
}

/**
 * It prints on the serial interface a buffer
 * @param[in] d buffer data
 * @param[in] count number of 32bit words to show
 */
void trtl_fw_print_data(uint32_t *d, unsigned int count)
{
	int i;

	if (!HAS_MOCKTURTLE_FRAMEWORK_DEBUG_ENABLE) {
		return;
	}

	for (i = 0; i < count; i++) {
		pr_debug("%s: data[%d] = 0x%"PRIx32"\n\r", __func__, i , d[i]);
		delay(1000);
	}
}


/**
 * It prints on the serial console the given message
 * @param[in] msg a mock turtle message
 */
void trtl_fw_print_message(struct trtl_msg *msg)
{
	pr_debug("Message\n\r");
	trtl_fw_print_header(msg->header);
	trtl_fw_print_data(msg->payload, msg->header->len);
}


/**
 * It get the current time from the internal WRNC timer
 * @param[out] seconds
 * @param[out] cycles
 */
void trtl_fw_time(uint32_t *seconds, uint32_t *cycles)
{
	*seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	*cycles = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
}


/**
 * It validates the user application actions
 * @param[in] app the user application
 * @return 0 on success. -EINVAL on error
 */
static inline int trtl_fw_init_action(struct trtl_fw_application *app)
{
	int i;

	/* Validate actions */
	for (i = 0; i < _app->n_actions; ++i) {
		if (i > __TRTL_MSG_ID_MAX_USER) {
			pr_error("Too many actions defined. Maximum '%d'\n\r",
				  __TRTL_MSG_ID_MAX_USER);
			return -EINVAL;
		}
		if (_app->actions[i] == NULL) {
			pr_error("Undefined action %d\n\r", i);
			return -EINVAL;
		}
	}
	return 0;
}


/**
 * It validates the FPGA
 * @param[in] app the user application
 * @return 0 on success. -EINVAL on error
 */
static inline int trtl_fw_init_fpga(struct trtl_fw_application *app)
{
	int i, found = 0;

	for (i = 0; i < app->fpga_id_compat_n; i++) {
		if (app->fpga_id_compat[i] == app->cfgrom->app_id) {
			found = 1;
			break;
		}
	}

	if (!found) {
		pr_error("\tFPGA '0x%"PRIx32"' not compatible with RT app: 0x%"PRIx32" 0x%"PRIx32" (git %"PRIx32")\n\r",
			 app->cfgrom->app_id,
			 app->version.rt_id,
			 app->version.rt_version,
			 app->version.git_version);
		return -EINVAL;
	}

	pr_debug("Running application '%s'\n\r", app->name);
	pr_debug("\tapp id     \t'0x%"PRIx32"'\n\r",
		 app->version.rt_id);
	pr_debug("\tapp version\t'%"PRId32".%"PRId32"'\n\r",
		 RT_VERSION_MAJ(app->version.rt_version),
		 RT_VERSION_MIN(app->version.rt_version));
	pr_debug("\tsource id  \t'0x%"PRIx32"'\n\r",
		 app->version.git_version);

	return 0;
}


/**
 * It validates the RMQ and the HMQ
 * @param[in] app the user application
 * @return 0 on success. -EINVAL on error
 */
static inline int trtl_fw_init_mq(struct trtl_fw_application *app)
{
	return 0;
}


/**
 * It validates the RMQ and the HMQ
 * @param[in] app the user application
 * @return 0 on success. -EINVAL on error
 */
static inline int trtl_fw_init_variables(struct trtl_fw_application *app)
{
	int i;

	pr_debug("Exported Variables\n\r");
	for (i = 0; i < app->n_variables; ++i)
		pr_debug("[%d] addr %p | mask 0x%"PRIx32" | off %d\n\r",
			  i,
			  app->variables[i].addr,
			  app->variables[i].mask,
			  app->variables[i].offset);

	return 0;
}


/**
 * It validates the RMQ and the HMQ
 * @param[in] app the user application
 * @return 0 on success. -EINVAL on error
 */
static inline int trtl_fw_init_buffers(struct trtl_fw_application *app)
{
	int i;

	pr_debug("Exported Buffers\n\r");
	for (i = 0; i < app->n_buffers; ++i)
		pr_debug("[%d] addr %p | len %"PRId32"\n\r", i,
			  app->buffers[i].buf,
			  app->buffers[i].len);

	return 0;
}


/**
 * The function initializes the library and it does some compatibility check.
 *
 * Initialization
 * - purge all the output host-message-queue
 *
 * compatibility checks:
 * - check bitstream FPGA ID if compatibile with the application. The value in
 *   app->version.fpga_id will be updated with the FPGA ID from the bitstream
 *
 * @param[in] app application declaration
 * @return 0 on success. -1 on error
 */
int trtl_fw_init(struct trtl_fw_application *app)
{
	int err;

	_app = app;
	_app->seq = 0;
	_app->cfgrom = trtl_config_rom_get();

	err = trtl_fw_init_action(_app);
	if (err)
		return err;

	err = trtl_fw_init_fpga(_app);
	if (err)
		return err;

	err = trtl_fw_init_mq(_app);
	if (err)
		return err;

	err = trtl_fw_init_buffers(_app);
	if (err)
		return err;

	err = trtl_fw_init_variables(_app);
	if (err)
		return err;

	return 0;
}
