import os
import PyMockTurtle
import pytest
import serial
import time

@pytest.fixture
def trtl_binary_hmq_purge(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-purge/fw-hmq-purge.bin")

@pytest.fixture
def trtl_binary_hmq_async_send(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-async-send/fw-hmq-async-send.bin")

@pytest.fixture
def trtl_binary_hmq_async_recv(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-async-recv/fw-hmq-async-recv.bin")

@pytest.fixture
def trtl_binary_hmq_sync(trtl_firmware_dir):
    return os.path.join(trtl_firmware_dir,
                        "firmware/hmq-sync/fw-hmq-sync.bin")


class TestHmq(object):
    confirm = 'OK\r\n'

    def test_purge(self, trtl_cpu, trtl_binary_hmq_purge):
        trtl_cpu.load_application_file(trtl_binary_hmq_purge)

        # fill HMQ
        for hmq in trtl_cpu.hmq:
            hmq.flush()
            msg = PyMockTurtle.TrtlMessage()
            msg.header.len = 1
            msg.payload[0] = 0x55555555
            hmq.send_msg(msg)

        # Run and check
        filename = "/dev/ttytrtl-{:04x}-{:d}".format(trtl_cpu.trtl_dev.device_id,
                                                     trtl_cpu.idx_cpu)
        with serial.Serial(filename) as ser:
            ser.reset_input_buffer()
            trtl_cpu.enable()
            time.sleep(1)
            assert ser.in_waiting >= len(self.confirm)
            assert self.confirm == ser.read(ser.in_waiting).decode()

    def test_purge_overflow(self, trtl_cpu, trtl_binary_hmq_purge):
        trtl_cpu.load_application_file(trtl_binary_hmq_purge)

        # Fill HMQ
        for hmq in trtl_cpu.hmq:
            idx_cfg = trtl_cpu.idx_cpu * PyMockTurtle.TRTL_CONFIG_ROM_MAX_CPU + hmq.idx_hmq
            # write more that what the HMQ can store
            for n in range(trtl_cpu.trtl_dev.rom.hmq[idx_cfg].entries):
                msg = PyMockTurtle.TrtlMessage()
                msg.header.len = 1
                msg.payload[0] = 0xAAAAAAAA
                hmq.send_msg(msg)

        # Run and check
        filename = "/dev/ttytrtl-{:04x}-{:d}".format(trtl_cpu.trtl_dev.device_id,
                                                     trtl_cpu.idx_cpu)
        with serial.Serial(filename) as ser:
            ser.reset_input_buffer()
            trtl_cpu.enable()
            time.sleep(1)
            assert ser.in_waiting >= len(self.confirm)
            assert self.confirm == ser.read(ser.in_waiting).decode()

    def test_async_send(self, trtl_cpu, trtl_msg, trtl_binary_hmq_async_send):
        """It sends the test messages on all available HMQ.
        The test is successful when we read "OK" from the serial console
        """
        if trtl_cpu.idx_cpu != 0:
            return
        trtl_cpu.load_application_file(trtl_binary_hmq_async_send)

        for hmq in trtl_cpu.hmq:
            hmq.flush()
            for msg in trtl_msg:
                hmq.send_msg(msg)
        filename = "/dev/ttytrtl-{:04x}-{:d}".format(trtl_cpu.trtl_dev.device_id,
                                                     trtl_cpu.idx_cpu)
        with serial.Serial(filename) as ser:
            ser.reset_input_buffer()
            trtl_cpu.enable()
            time.sleep(1)
            assert ser.in_waiting >= len(self.confirm)
            assert self.confirm == ser.read(ser.in_waiting).decode()

    def test_async_recv(self, trtl_cpu, trtl_binary_hmq_async_recv):
        """It receivess messages on all available HMQ.
        The test is successful when we recevie all the expected messages
        """
        if trtl_cpu.idx_cpu != 0:
            return
        trtl_cpu.load_application_file(trtl_binary_hmq_async_recv)

        for hmq in trtl_cpu.hmq:
            hmq.flush()

        trtl_cpu.enable()
        time.sleep(0.1)

        for hmq in trtl_cpu.hmq:
            idx_cfg = trtl_cpu.idx_cpu * PyMockTurtle.TRTL_CONFIG_ROM_MAX_CPU + hmq.idx_hmq
            for n in range(trtl_cpu.trtl_dev.rom.hmq[idx_cfg].entries):
                msg = hmq.recv_msg()
                payload = list(msg.payload)
                for i, val in enumerate(payload):
                    if i >= msg.header.len:
                        break
                    assert i == val

    def test_sync(self, trtl_cpu, trtl_msg, trtl_binary_hmq_sync):
        """It sends the test messages on all available HMQ.
        The test is successful when what we read back is the same we sent
        """
        if trtl_cpu.idx_cpu != 0:
            return

        trtl_cpu.load_application_file(trtl_binary_hmq_sync)
        trtl_cpu.enable()
        time.sleep(0.1)

        for hmq in trtl_cpu.hmq:
            hmq.flush()

            if hmq.idx_hmq != 0:
                continue

            for msg in trtl_msg:
                print(msg.payload[0], msg.payload[1])
                msg_r = hmq.sync_msg(msg)
                # the library add automatically this flag, but for the purpose
                # of this test assertiong we manually set it
                msg.header.flags |= PyMockTurtle.TrtlHmqHeader.TRTL_HMQ_HEADER_FLAG_SYNC
                assert msg_r == msg
